"""SDK configuration."""

from __future__ import annotations

import os
from dataclasses import dataclass, field


_DEFAULT_BASE_URL = "https://api.melogen.ai"
_DEFAULT_TIMEOUT = 30.0
_DEFAULT_POLL_INTERVAL = 2.0
_DEFAULT_POLL_TIMEOUT = 300.0


@dataclass(frozen=True)
class Config:
    """Immutable configuration for the Melogen client."""

    api_key: str
    base_url: str = field(default_factory=lambda: os.getenv("MELOGEN_BASE_URL", _DEFAULT_BASE_URL))
    timeout: float = _DEFAULT_TIMEOUT
    poll_interval: float = _DEFAULT_POLL_INTERVAL
    poll_timeout: float = _DEFAULT_POLL_TIMEOUT

    def __post_init__(self) -> None:
        if not self.api_key:
            raise ValueError(
                "API key is required. Pass api_key= or set MELOGEN_API_KEY env var."
            )

    @classmethod
    def from_env(cls) -> Config:
        api_key = os.getenv("MELOGEN_API_KEY", "")
        return cls(api_key=api_key)



"""Music analysis client with SSE streaming support."""

from __future__ import annotations

from collections.abc import AsyncGenerator, Generator
from typing import Any

from melogenai._http import AsyncHttpClient, HttpClient
from melogenai.types import AnalysisResult

SUPPORTED_LANGUAGES = frozenset(
    {"en", "zh", "tw", "ja", "ko", "de", "fr", "es", "pt", "it", "ru", "pl", "el", "vi"}
)


def _build_payload(
    mxl_file_url: str | None,
    image_url: str | None,
    pdf_file_url: str | None,
    language: str,
) -> dict[str, Any]:
    if not any([mxl_file_url, image_url, pdf_file_url]):
        raise ValueError(
            "At least one of mxl_file_url, image_url, or pdf_file_url must be provided."
        )
    if language not in SUPPORTED_LANGUAGES:
        raise ValueError(
            f"Unsupported language '{language}'. Supported: {', '.join(sorted(SUPPORTED_LANGUAGES))}"
        )

    payload: dict[str, Any] = {"lang": language}
    if mxl_file_url:
        payload["mxl_file_url"] = mxl_file_url
    if image_url:
        payload["image_url"] = image_url
    if pdf_file_url:
        payload["pdf_file_url"] = pdf_file_url
    return payload


class AnalysisClient:
    """Synchronous client for music analysis."""

    def __init__(self, http: HttpClient):
        self._http = http

    def analyze(
        self,
        *,
        mxl_file_url: str | None = None,
        image_url: str | None = None,
        pdf_file_url: str | None = None,
        language: str = "en",
        timeout: float | None = None,
    ) -> AnalysisResult:
        """Analyze music structure, returning the full result.

        At least one of mxl_file_url, image_url, or pdf_file_url must be provided.

        Args:
            mxl_file_url: URL of a MusicXML (.mxl) file.
            image_url: URL of a sheet music image.
            pdf_file_url: URL of a PDF score.
            language: Language for analysis results (e.g. "en", "zh", "ja").
            timeout: Request timeout in seconds.
        """
        payload = _build_payload(mxl_file_url, image_url, pdf_file_url, language)
        text = self._http.consume_sse("/api/analysis/analyze", payload, timeout=timeout or 300)
        return AnalysisResult(status="completed", analysis=text, language=language)

    def analyze_stream(
        self,
        *,
        mxl_file_url: str | None = None,
        image_url: str | None = None,
        pdf_file_url: str | None = None,
        language: str = "en",
        timeout: float | None = None,
    ) -> Generator[str, None, None]:
        """Stream analysis results chunk by chunk."""
        payload = _build_payload(mxl_file_url, image_url, pdf_file_url, language)
        yield from self._http.stream_sse("/api/analysis/analyze", payload, timeout=timeout or 300)


class AsyncAnalysisClient:
    """Asynchronous client for music analysis."""

    def __init__(self, http: AsyncHttpClient):
        self._http = http

    async def analyze(
        self,
        *,
        mxl_file_url: str | None = None,
        image_url: str | None = None,
        pdf_file_url: str | None = None,
        language: str = "en",
        timeout: float | None = None,
    ) -> AnalysisResult:
        """Analyze music structure, returning the full result."""
        payload = _build_payload(mxl_file_url, image_url, pdf_file_url, language)
        text = await self._http.consume_sse(
            "/api/analysis/analyze", payload, timeout=timeout or 300
        )
        return AnalysisResult(status="completed", analysis=text, language=language)

    async def analyze_stream(
        self,
        *,
        mxl_file_url: str | None = None,
        image_url: str | None = None,
        pdf_file_url: str | None = None,
        language: str = "en",
        timeout: float | None = None,
    ) -> AsyncGenerator[str, None]:
        """Stream analysis results chunk by chunk."""
        payload = _build_payload(mxl_file_url, image_url, pdf_file_url, language)
        async for chunk in self._http.stream_sse(
            "/api/analysis/analyze", payload, timeout=timeout or 300
        ):
            yield chunk



"""Shared types for the Melogen SDK."""

from typing import Any, Dict, Optional

from pydantic import BaseModel, Field


class TaskResult(BaseModel):
    """Result of a conversion task submission."""

    task_id: str
    status: str = "pending"


class TaskStatus(BaseModel):
    """Full status of a task including download URLs."""

    task_id: str
    status: str
    download_urls: Dict[str, str] = Field(default_factory=dict)
    error: Optional[str] = None
    task: Dict[str, Any] = Field(default_factory=dict)


class AnalysisResult(BaseModel):
    """Result of a music analysis."""

    status: str
    analysis: str
    language: str


COMPLETED_STATUSES = frozenset({"completed", "done", "finished", "success"})
FAILED_STATUSES = frozenset({"failed", "error", "cancelled"})



"""Shared test fixtures."""

from __future__ import annotations

import pytest
import respx

from melogenai.auth import AuthManager
from melogenai.config import Config


TEST_API_KEY = "melo_test_key_123"
TEST_BASE_URL = "https://api.test.melogen.ai"
TEST_JWT = "eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiJ9.test"


@pytest.fixture
def config() -> Config:
    return Config(api_key=TEST_API_KEY, base_url=TEST_BASE_URL)


@pytest.fixture
def auth_manager(config: Config) -> AuthManager:
    return AuthManager(config)


@pytest.fixture
def mock_api():
    """Provide a respx mock router scoped to the test base URL."""
    with respx.mock(base_url=TEST_BASE_URL) as router:
        # Always mock the auth endpoint
        router.post("/api/api-keys/verify", name="auth_verify").respond(
            json={
                "valid": True,
                "access_token": TEST_JWT,
                "user_id": "user_123",
                "scopes": ["sheet2midi", "music2midi", "analysis"],
            }
        )
        yield router



"""Tests for music2midi service client."""

from __future__ import annotations

import json

import pytest
import respx
from httpx import Response

from melogenai.client import AsyncMelogen, Melogen

from conftest import TEST_API_KEY, TEST_BASE_URL


class TestMusic2MidiSync:
    def test_convert(self, mock_api: respx.MockRouter) -> None:
        mock_api.post("/api/music2midi/convert").respond(
            json={"task_id": "task_101", "status": "pending"}
        )

        client = Melogen(api_key=TEST_API_KEY, base_url=TEST_BASE_URL)
        result = client.music2midi.convert("https://example.com/song.mp3")

        assert result.task_id == "task_101"

    def test_convert_with_options(self, mock_api: respx.MockRouter) -> None:
        route = mock_api.post("/api/music2midi/convert").respond(
            json={"task_id": "task_102", "status": "pending"}
        )

        client = Melogen(api_key=TEST_API_KEY, base_url=TEST_BASE_URL)
        result = client.music2midi.convert(
            "https://example.com/song.mp3", save_stems=True, quantize=True
        )

        assert result.task_id == "task_102"
        request = route.calls.last.request
        body = json.loads(request.content)
        assert body["save_stems"] is True
        assert body["quantize"] is True

    def test_get_task(self, mock_api: respx.MockRouter) -> None:
        mock_api.get("/api/music2midi/tasks/task_101").respond(
            json={
                "task": {"status": "completed"},
                "download_urls": {"midi": "https://dl.example.com/out.midi"},
            }
        )

        client = Melogen(api_key=TEST_API_KEY, base_url=TEST_BASE_URL)
        status = client.music2midi.get_task("task_101")

        assert status.status == "completed"

    def test_wait_completes(self, mock_api: respx.MockRouter) -> None:
        mock_api.get("/api/music2midi/tasks/task_101").mock(
            side_effect=[
                Response(200, json={"task": {"status": "processing"}, "download_urls": {}}),
                Response(
                    200,
                    json={
                        "task": {"status": "completed"},
                        "download_urls": {"midi": "https://dl.example.com/out.midi"},
                    },
                ),
            ]
        )

        client = Melogen(api_key=TEST_API_KEY, base_url=TEST_BASE_URL, poll_interval=0.01)
        result = client.music2midi.wait("task_101")

        assert result.status == "completed"


class TestMusic2MidiAsync:
    @pytest.mark.asyncio
    async def test_convert(self, mock_api: respx.MockRouter) -> None:
        mock_api.post("/api/music2midi/convert").respond(
            json={"task_id": "task_101", "status": "pending"}
        )

        async with AsyncMelogen(api_key=TEST_API_KEY, base_url=TEST_BASE_URL) as client:
            result = await client.music2midi.convert("https://example.com/song.mp3")

        assert result.task_id == "task_101"



"""Sheet music to MIDI conversion client."""

from __future__ import annotations

from typing import Any

from melogenai._http import AsyncHttpClient, HttpClient
from melogenai.types import TaskResult, TaskStatus


class Sheet2MidiClient:
    """Synchronous client for sheet music to MIDI conversion."""

    def __init__(self, http: HttpClient):
        self._http = http

    def convert(
        self,
        file_url: str,
        *,
        output_formats: list[str] | None = None,
        convert_midi: bool = True,
    ) -> TaskResult:
        """Submit a sheet music file for conversion.

        Args:
            file_url: URL of the sheet music file (PDF, PNG, or JPG).
            output_formats: Desired output formats, e.g. ["midi", "musicxml"].
            convert_midi: Whether to generate MIDI output.

        Returns:
            TaskResult with task_id and initial status.
        """
        payload: dict[str, Any] = {
            "file_url": file_url,
            "convert_midi": convert_midi,
        }
        if output_formats is not None:
            payload["output_formats"] = output_formats

        data = self._http.post("/api/sheet2midi/convert", payload, timeout=60)
        return TaskResult(task_id=data["task_id"], status=data.get("status", "pending"))

    def get_task(self, task_id: str) -> TaskStatus:
        """Get the current status of a conversion task."""
        data = self._http.get(f"/api/sheet2midi/tasks/{task_id}")
        task = data.get("task", data)
        return TaskStatus(
            task_id=task_id,
            status=task.get("status", "unknown"),
            download_urls=data.get("download_urls", {}),
            error=task.get("error"),
            task=task,
        )

    def wait(
        self,
        task_id: str,
        *,
        poll_interval: float | None = None,
        timeout: float | None = None,
    ) -> TaskStatus:
        """Poll until the task completes and return the final status."""
        data = self._http.poll_task(
            f"/api/sheet2midi/tasks/{task_id}",
            task_id,
            interval=poll_interval,
            timeout=timeout,
        )
        task = data.get("task", data)
        return TaskStatus(
            task_id=task_id,
            status=task.get("status", "completed"),
            download_urls=data.get("download_urls", {}),
            task=task,
        )


class AsyncSheet2MidiClient:
    """Asynchronous client for sheet music to MIDI conversion."""

    def __init__(self, http: AsyncHttpClient):
        self._http = http

    async def convert(
        self,
        file_url: str,
        *,
        output_formats: list[str] | None = None,
        convert_midi: bool = True,
    ) -> TaskResult:
        """Submit a sheet music file for conversion."""
        payload: dict[str, Any] = {
            "file_url": file_url,
            "convert_midi": convert_midi,
        }
        if output_formats is not None:
            payload["output_formats"] = output_formats

        data = await self._http.post("/api/sheet2midi/convert", payload, timeout=60)
        return TaskResult(task_id=data["task_id"], status=data.get("status", "pending"))

    async def get_task(self, task_id: str) -> TaskStatus:
        """Get the current status of a conversion task."""
        data = await self._http.get(f"/api/sheet2midi/tasks/{task_id}")
        task = data.get("task", data)
        return TaskStatus(
            task_id=task_id,
            status=task.get("status", "unknown"),
            download_urls=data.get("download_urls", {}),
            error=task.get("error"),
            task=task,
        )

    async def wait(
        self,
        task_id: str,
        *,
        poll_interval: float | None = None,
        timeout: float | None = None,
    ) -> TaskStatus:
        """Poll until the task completes and return the final status."""
        data = await self._http.poll_task(
            f"/api/sheet2midi/tasks/{task_id}",
            task_id,
            interval=poll_interval,
            timeout=timeout,
        )
        task = data.get("task", data)
        return TaskStatus(
            task_id=task_id,
            status=task.get("status", "completed"),
            download_urls=data.get("download_urls", {}),
            task=task,
        )



"""Main Melogen client classes."""

from __future__ import annotations

import os
from types import TracebackType
from typing import Any

from melogenai._http import AsyncHttpClient, HttpClient
from melogenai.analysis.client import AnalysisClient, AsyncAnalysisClient
from melogenai.auth import AuthManager
from melogenai.config import Config
from melogenai.music2midi.client import AsyncMusic2MidiClient, Music2MidiClient
from melogenai.sheet2midi.client import AsyncSheet2MidiClient, Sheet2MidiClient


class Melogen:
    """Synchronous Melogen client.

    Usage::

        client = Melogen(api_key="melo_xxx")
        result = client.sheet2midi.convert(file_url="https://...")
        final = client.sheet2midi.wait(result.task_id)
    """

    def __init__(
        self,
        *,
        api_key: str | None = None,
        base_url: str | None = None,
        timeout: float = 30.0,
        poll_interval: float = 2.0,
        poll_timeout: float = 300.0,
    ):
        resolved_key = api_key or os.getenv("MELOGEN_API_KEY", "")
        config_kwargs: dict[str, Any] = {
            "api_key": resolved_key,
            "timeout": timeout,
            "poll_interval": poll_interval,
            "poll_timeout": poll_timeout,
        }
        if base_url is not None:
            config_kwargs["base_url"] = base_url

        self._config = Config(**config_kwargs)
        self._auth = AuthManager(self._config)
        self._http = HttpClient(self._config, self._auth)

        self.sheet2midi = Sheet2MidiClient(self._http)
        self.music2midi = Music2MidiClient(self._http)
        self.analysis = AnalysisClient(self._http)


class AsyncMelogen:
    """Asynchronous Melogen client.

    Usage::

        async with AsyncMelogen(api_key="melo_xxx") as client:
            result = await client.sheet2midi.convert(file_url="https://...")
            final = await client.sheet2midi.wait(result.task_id)
    """

    def __init__(
        self,
        *,
        api_key: str | None = None,
        base_url: str | None = None,
        timeout: float = 30.0,
        poll_interval: float = 2.0,
        poll_timeout: float = 300.0,
    ):
        resolved_key = api_key or os.getenv("MELOGEN_API_KEY", "")
        config_kwargs: dict[str, Any] = {
            "api_key": resolved_key,
            "timeout": timeout,
            "poll_interval": poll_interval,
            "poll_timeout": poll_timeout,
        }
        if base_url is not None:
            config_kwargs["base_url"] = base_url

        self._config = Config(**config_kwargs)
        self._auth = AuthManager(self._config)
        self._http = AsyncHttpClient(self._config, self._auth)

        self.sheet2midi = AsyncSheet2MidiClient(self._http)
        self.music2midi = AsyncMusic2MidiClient(self._http)
        self.analysis = AsyncAnalysisClient(self._http)

    async def __aenter__(self) -> AsyncMelogen:
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        pass



"""Tests for auth module."""

from __future__ import annotations

import pytest
import respx
from httpx import Response

from melogenai.auth import AuthManager
from melogenai.config import Config
from melogenai.exceptions import AuthenticationError

from conftest import TEST_API_KEY, TEST_BASE_URL, TEST_JWT


class TestAuthManager:
    def test_get_token_success(self, config: Config, mock_api: respx.MockRouter) -> None:
        auth = AuthManager(config)
        token_data = auth.get_token(TEST_API_KEY)

        assert token_data["token"] == TEST_JWT
        assert token_data["user_id"] == "user_123"
        assert "expires_at" in token_data

    def test_get_token_cached(self, config: Config, mock_api: respx.MockRouter) -> None:
        auth = AuthManager(config)

        # First call hits the API
        token_data1 = auth.get_token(TEST_API_KEY)
        # Second call should use cache
        token_data2 = auth.get_token(TEST_API_KEY)

        assert token_data1["token"] == token_data2["token"]
        assert mock_api["auth_verify"].call_count == 1

    def test_get_token_invalid_key(self, config: Config, mock_api: respx.MockRouter) -> None:
        mock_api.post("/api/api-keys/verify").respond(
            json={"valid": False},
        )
        auth = AuthManager(config)

        with pytest.raises(AuthenticationError, match="Invalid API key"):
            auth.get_token(TEST_API_KEY)

    def test_invalidate_clears_cache(self, config: Config, mock_api: respx.MockRouter) -> None:
        auth = AuthManager(config)
        auth.get_token(TEST_API_KEY)

        auth.invalidate(TEST_API_KEY)

        # Next call should hit the API again
        auth.get_token(TEST_API_KEY)
        assert mock_api["auth_verify"].call_count == 2

    @pytest.mark.asyncio
    async def test_async_get_token_success(self, config: Config, mock_api: respx.MockRouter) -> None:
        auth = AuthManager(config)
        token_data = await auth.async_get_token(TEST_API_KEY)

        assert token_data["token"] == TEST_JWT
        assert token_data["user_id"] == "user_123"

    @pytest.mark.asyncio
    async def test_async_get_token_cached(self, config: Config, mock_api: respx.MockRouter) -> None:
        auth = AuthManager(config)
        await auth.async_get_token(TEST_API_KEY)
        await auth.async_get_token(TEST_API_KEY)

        assert mock_api["auth_verify"].call_count == 1



"""Basic conversion examples using the Melogen SDK."""

import os

from melogenai import Melogen, AuthenticationError, TaskTimeoutError


def main():
    api_key = os.getenv("MELOGEN_API_KEY")
    if not api_key:
        print("Set MELOGEN_API_KEY environment variable first")
        return

    client = Melogen(api_key=api_key)

    # --- Sheet music to MIDI ---
    print("Converting sheet music to MIDI...")
    try:
        result = client.sheet2midi.convert(
            file_url="https://example.com/score.pdf",
            output_formats=["midi", "musicxml"],
        )
        print(f"Task created: {result.task_id}")

        final = client.sheet2midi.wait(result.task_id)
        print(f"Status: {final.status}")
        for name, url in final.download_urls.items():
            print(f"  {name}: {url}")
    except AuthenticationError:
        print("Invalid API key")
    except TaskTimeoutError:
        print("Task timed out")

    # --- Audio to MIDI ---
    print("\nConverting audio to MIDI...")
    try:
        result = client.music2midi.convert(
            file_url="https://example.com/song.mp3",
            save_stems=True,
            quantize=True,
        )
        print(f"Task created: {result.task_id}")

        final = client.music2midi.wait(result.task_id)
        print(f"Status: {final.status}")
        for name, url in final.download_urls.items():
            print(f"  {name}: {url}")
    except AuthenticationError:
        print("Invalid API key")
    except TaskTimeoutError:
        print("Task timed out")

    # --- Music analysis (streaming) ---
    print("\nAnalyzing music structure...")
    try:
        for chunk in client.analysis.analyze_stream(
            image_url="https://example.com/score.png",
            language="en",
        ):
            print(chunk, end="", flush=True)
        print()
    except AuthenticationError:
        print("Invalid API key")


if __name__ == "__main__":
    main()



"""Tests for CLI commands."""

from __future__ import annotations

import pytest
import respx
from click.testing import CliRunner
from httpx import Response

from melogenai.cli.main import cli

from conftest import TEST_API_KEY, TEST_BASE_URL


@pytest.fixture
def runner():
    return CliRunner()


class TestCLI:
    def test_version(self, runner: CliRunner) -> None:
        result = runner.invoke(cli, ["--version"])
        assert result.exit_code == 0
        assert "melogen" in result.output

    def test_sheet2midi_command(self, runner: CliRunner, mock_api: respx.MockRouter) -> None:
        mock_api.post("/api/sheet2midi/convert").respond(
            json={"task_id": "task_001", "status": "pending"}
        )
        mock_api.get("/api/sheet2midi/tasks/task_001").mock(
            side_effect=[
                Response(
                    200,
                    json={
                        "task": {"status": "completed"},
                        "download_urls": {"midi": "https://dl.example.com/out.midi"},
                    },
                ),
            ]
        )

        result = runner.invoke(
            cli,
            ["--api-key", TEST_API_KEY, "--base-url", TEST_BASE_URL, "sheet2midi", "https://example.com/score.pdf"],
        )

        assert result.exit_code == 0
        assert "completed" in result.output.lower() or "midi" in result.output.lower()

    def test_music2midi_command(self, runner: CliRunner, mock_api: respx.MockRouter) -> None:
        mock_api.post("/api/music2midi/convert").respond(
            json={"task_id": "task_101", "status": "pending"}
        )
        mock_api.get("/api/music2midi/tasks/task_101").mock(
            side_effect=[
                Response(
                    200,
                    json={
                        "task": {"status": "completed"},
                        "download_urls": {"midi": "https://dl.example.com/out.midi"},
                    },
                ),
            ]
        )

        result = runner.invoke(
            cli,
            ["--api-key", TEST_API_KEY, "--base-url", TEST_BASE_URL, "music2midi", "https://example.com/song.mp3"],
        )

        assert result.exit_code == 0

    def test_analysis_requires_input(self, runner: CliRunner) -> None:
        result = runner.invoke(
            cli,
            ["--api-key", TEST_API_KEY, "--base-url", TEST_BASE_URL, "analysis"],
        )

        assert result.exit_code != 0
        assert "provide at least one" in result.output.lower()

    def test_auth_verify(self, runner: CliRunner, mock_api: respx.MockRouter) -> None:
        result = runner.invoke(
            cli,
            ["--api-key", TEST_API_KEY, "--base-url", TEST_BASE_URL, "auth", "verify"],
        )

        assert result.exit_code == 0
        assert "valid" in result.output.lower()



"""Low-level HTTP client wrapper for sync and async usage."""

from __future__ import annotations

import asyncio
import json
import time
from collections.abc import Generator, AsyncGenerator
from typing import Any

import httpx

from melogenai.auth import AuthManager
from melogenai.config import Config
from melogenai.exceptions import (
    AuthenticationError,
    InsufficientCreditsError,
    MelogenError,
    RateLimitError,
    TaskFailedError,
    TaskTimeoutError,
    ValidationError,
)
from melogenai.types import COMPLETED_STATUSES, FAILED_STATUSES


def _raise_for_status(resp: httpx.Response) -> None:
    """Map HTTP status codes to SDK exceptions."""
    if resp.is_success:
        return
    status = resp.status_code
    try:
        body = resp.json()
        detail = body.get("detail") or body.get("error") or body.get("message") or resp.text
    except Exception:
        detail = resp.text

    if status == 401:
        raise AuthenticationError(str(detail))
    if status in (402, 403):
        raise InsufficientCreditsError(str(detail))
    if status == 422:
        raise ValidationError(str(detail))
    if status == 429:
        raise RateLimitError(str(detail))
    raise MelogenError(f"HTTP {status}: {detail}", status_code=status)


class HttpClient:
    """Synchronous HTTP client."""

    def __init__(self, config: Config, auth: AuthManager):
        self._config = config
        self._auth = auth

    def _headers(self, token: str) -> dict[str, str]:
        return {
            "Authorization": f"Bearer {token}",
            "Content-Type": "application/json",
        }

    def _get_token(self) -> str:
        token_data = self._auth.get_token(self._config.api_key)
        return token_data["token"]

    def request(
        self,
        method: str,
        path: str,
        *,
        json_body: dict[str, Any] | None = None,
        timeout: float | None = None,
    ) -> dict[str, Any]:
        token = self._get_token()
        with httpx.Client(timeout=timeout or self._config.timeout) as client:
            resp = client.request(
                method,
                f"{self._config.base_url}{path}",
                json=json_body,
                headers=self._headers(token),
            )
        if resp.status_code == 401:
            self._auth.invalidate(self._config.api_key)
        _raise_for_status(resp)
        return resp.json()

    def post(self, path: str, payload: dict[str, Any], timeout: float | None = None) -> dict[str, Any]:
        return self.request("POST", path, json_body=payload, timeout=timeout)

    def get(self, path: str, timeout: float | None = None) -> dict[str, Any]:
        return self.request("GET", path, timeout=timeout)

    def poll_task(
        self,
        path: str,
        task_id: str,
        interval: float | None = None,
        timeout: float | None = None,
    ) -> dict[str, Any]:
        interval = interval or self._config.poll_interval
        timeout = timeout or self._config.poll_timeout
        deadline = time.monotonic() + timeout

        while time.monotonic() < deadline:
            result = self.get(path)
            task = result.get("task", result)
            status = task.get("status", "")

            if status in COMPLETED_STATUSES:
                return result
            if status in FAILED_STATUSES:
                error_msg = task.get("error") or task.get("message") or "Task failed"
                raise TaskFailedError(task_id, error_msg)

            time.sleep(interval)

        raise TaskTimeoutError(task_id, timeout)

    def stream_sse(
        self,
        path: str,
        payload: dict[str, Any],
        timeout: float | None = None,
    ) -> Generator[str, None, None]:
        """POST to SSE endpoint and yield content chunks."""
        token = self._get_token()
        t = timeout or self._config.poll_timeout
        with httpx.Client(timeout=httpx.Timeout(t, connect=10)) as client:
            with client.stream(
                "POST",
                f"{self._config.base_url}{path}",
                json=payload,
                headers=self._headers(token),
            ) as resp:
                if resp.status_code == 401:
                    self._auth.invalidate(self._config.api_key)
                _raise_for_status(resp)

                for line in resp.iter_lines():
                    if not line.startswith("data: "):
                        continue
                    data_str = line[6:].strip()
                    if data_str == "[DONE]":
                        return

                    try:
                        event = json.loads(data_str)
                    except json.JSONDecodeError:
                        continue

                    chunk = _extract_sse_content(event)
                    if chunk is not None:
                        yield chunk

    def consume_sse(
        self,
        path: str,
        payload: dict[str, Any],
        timeout: float | None = None,
    ) -> str:
        """POST to SSE endpoint and return the full aggregated content."""
        return "".join(self.stream_sse(path, payload, timeout))


class AsyncHttpClient:
    """Asynchronous HTTP client."""

    def __init__(self, config: Config, auth: AuthManager):
        self._config = config
        self._auth = auth

    def _headers(self, token: str) -> dict[str, str]:
        return {
            "Authorization": f"Bearer {token}",
            "Content-Type": "application/json",
        }

    async def _get_token(self) -> str:
        token_data = await self._auth.async_get_token(self._config.api_key)
        return token_data["token"]

    async def request(
        self,
        method: str,
        path: str,
        *,
        json_body: dict[str, Any] | None = None,
        timeout: float | None = None,
    ) -> dict[str, Any]:
        token = await self._get_token()
        async with httpx.AsyncClient(timeout=timeout or self._config.timeout) as client:
            resp = await client.request(
                method,
                f"{self._config.base_url}{path}",
                json=json_body,
                headers=self._headers(token),
            )
        if resp.status_code == 401:
            self._auth.invalidate(self._config.api_key)
        _raise_for_status(resp)
        return resp.json()

    async def post(self, path: str, payload: dict[str, Any], timeout: float | None = None) -> dict[str, Any]:
        return await self.request("POST", path, json_body=payload, timeout=timeout)

    async def get(self, path: str, timeout: float | None = None) -> dict[str, Any]:
        return await self.request("GET", path, timeout=timeout)

    async def poll_task(
        self,
        path: str,
        task_id: str,
        interval: float | None = None,
        timeout: float | None = None,
    ) -> dict[str, Any]:
        interval = interval or self._config.poll_interval
        timeout = timeout or self._config.poll_timeout
        deadline = time.monotonic() + timeout

        while time.monotonic() < deadline:
            result = await self.get(path)
            task = result.get("task", result)
            status = task.get("status", "")

            if status in COMPLETED_STATUSES:
                return result
            if status in FAILED_STATUSES:
                error_msg = task.get("error") or task.get("message") or "Task failed"
                raise TaskFailedError(task_id, error_msg)

            await asyncio.sleep(interval)

        raise TaskTimeoutError(task_id, timeout)

    async def stream_sse(
        self,
        path: str,
        payload: dict[str, Any],
        timeout: float | None = None,
    ) -> AsyncGenerator[str, None]:
        """POST to SSE endpoint and yield content chunks."""
        token = await self._get_token()
        t = timeout or self._config.poll_timeout
        async with httpx.AsyncClient(timeout=httpx.Timeout(t, connect=10)) as client:
            async with client.stream(
                "POST",
                f"{self._config.base_url}{path}",
                json=payload,
                headers=self._headers(token),
            ) as resp:
                if resp.status_code == 401:
                    self._auth.invalidate(self._config.api_key)
                _raise_for_status(resp)

                async for line in resp.aiter_lines():
                    if not line.startswith("data: "):
                        continue
                    data_str = line[6:].strip()
                    if data_str == "[DONE]":
                        return

                    try:
                        event = json.loads(data_str)
                    except json.JSONDecodeError:
                        continue

                    chunk = _extract_sse_content(event)
                    if chunk is not None:
                        yield chunk

    async def consume_sse(
        self,
        path: str,
        payload: dict[str, Any],
        timeout: float | None = None,
    ) -> str:
        """POST to SSE endpoint and return the full aggregated content."""
        chunks: list[str] = []
        async for chunk in self.stream_sse(path, payload, timeout):
            chunks.append(chunk)
        return "".join(chunks)


def _extract_sse_content(event: dict[str, Any]) -> str | None:
    """Extract content from an SSE event."""
    if event.get("status") in FAILED_STATUSES:
        error_msg = event.get("error") or event.get("message") or "Stream error"
        raise MelogenError(f"SSE stream error: {error_msg}")

    if "content" in event:
        return event["content"]
    if "text" in event:
        return event["text"]
    if "result" in event:
        result = event["result"]
        return result if isinstance(result, str) else json.dumps(result, ensure_ascii=False)
    return None



"""Tests for sheet2midi service client."""

from __future__ import annotations

import pytest
import respx
from httpx import Response

from melogenai.client import AsyncMelogen, Melogen

from conftest import TEST_API_KEY, TEST_BASE_URL


class TestSheet2MidiSync:
    def test_convert(self, mock_api: respx.MockRouter) -> None:
        mock_api.post("/api/sheet2midi/convert").respond(
            json={"task_id": "task_001", "status": "pending"}
        )

        client = Melogen(api_key=TEST_API_KEY, base_url=TEST_BASE_URL)
        result = client.sheet2midi.convert("https://example.com/score.pdf")

        assert result.task_id == "task_001"
        assert result.status == "pending"

    def test_convert_with_formats(self, mock_api: respx.MockRouter) -> None:
        route = mock_api.post("/api/sheet2midi/convert").respond(
            json={"task_id": "task_002", "status": "pending"}
        )

        client = Melogen(api_key=TEST_API_KEY, base_url=TEST_BASE_URL)
        result = client.sheet2midi.convert(
            "https://example.com/score.pdf",
            output_formats=["midi", "musicxml"],
            convert_midi=True,
        )

        assert result.task_id == "task_002"
        import json
        request = route.calls.last.request
        body = json.loads(request.content)
        assert body["output_formats"] == ["midi", "musicxml"]
        assert body["convert_midi"] is True

    def test_get_task(self, mock_api: respx.MockRouter) -> None:
        mock_api.get("/api/sheet2midi/tasks/task_001").respond(
            json={
                "task": {"status": "completed"},
                "download_urls": {"midi": "https://dl.example.com/out.midi"},
            }
        )

        client = Melogen(api_key=TEST_API_KEY, base_url=TEST_BASE_URL)
        status = client.sheet2midi.get_task("task_001")

        assert status.status == "completed"
        assert status.download_urls["midi"] == "https://dl.example.com/out.midi"

    def test_wait_completes(self, mock_api: respx.MockRouter) -> None:
        # First poll returns pending, second returns completed
        mock_api.get("/api/sheet2midi/tasks/task_001").mock(
            side_effect=[
                Response(200, json={"task": {"status": "pending"}, "download_urls": {}}),
                Response(
                    200,
                    json={
                        "task": {"status": "completed"},
                        "download_urls": {"midi": "https://dl.example.com/out.midi"},
                    },
                ),
            ]
        )

        client = Melogen(api_key=TEST_API_KEY, base_url=TEST_BASE_URL, poll_interval=0.01)
        result = client.sheet2midi.wait("task_001")

        assert result.status == "completed"
        assert "midi" in result.download_urls


class TestSheet2MidiAsync:
    @pytest.mark.asyncio
    async def test_convert(self, mock_api: respx.MockRouter) -> None:
        mock_api.post("/api/sheet2midi/convert").respond(
            json={"task_id": "task_001", "status": "pending"}
        )

        async with AsyncMelogen(api_key=TEST_API_KEY, base_url=TEST_BASE_URL) as client:
            result = await client.sheet2midi.convert("https://example.com/score.pdf")

        assert result.task_id == "task_001"

    @pytest.mark.asyncio
    async def test_wait_completes(self, mock_api: respx.MockRouter) -> None:
        mock_api.get("/api/sheet2midi/tasks/task_001").mock(
            side_effect=[
                Response(200, json={"task": {"status": "pending"}, "download_urls": {}}),
                Response(
                    200,
                    json={
                        "task": {"status": "completed"},
                        "download_urls": {"midi": "https://dl.example.com/out.midi"},
                    },
                ),
            ]
        )

        async with AsyncMelogen(
            api_key=TEST_API_KEY, base_url=TEST_BASE_URL, poll_interval=0.01
        ) as client:
            result = await client.sheet2midi.wait("task_001")

        assert result.status == "completed"



"""Exception hierarchy for the Melogen SDK."""

from __future__ import annotations


class MelogenError(Exception):
    """Base exception for all Melogen SDK errors."""

    def __init__(self, message: str, status_code: int | None = None):
        super().__init__(message)
        self.status_code = status_code


class AuthenticationError(MelogenError):
    """API key is invalid or expired."""

    def __init__(self, message: str = "Invalid or expired API key"):
        super().__init__(message, status_code=401)


class InsufficientCreditsError(MelogenError):
    """Account does not have enough credits."""

    def __init__(self, message: str = "Insufficient credits"):
        super().__init__(message, status_code=402)


class ValidationError(MelogenError):
    """Request parameters failed validation."""

    def __init__(self, message: str = "Validation error"):
        super().__init__(message, status_code=422)


class RateLimitError(MelogenError):
    """Too many requests."""

    def __init__(self, message: str = "Rate limit exceeded"):
        super().__init__(message, status_code=429)


class TaskTimeoutError(MelogenError):
    """Task polling exceeded the timeout."""

    def __init__(self, task_id: str, timeout: float):
        super().__init__(f"Task {task_id} timed out after {timeout}s")
        self.task_id = task_id
        self.timeout = timeout


class TaskFailedError(MelogenError):
    """Task completed with a failure status."""

    def __init__(self, task_id: str, error: str):
        super().__init__(f"Task {task_id} failed: {error}")
        self.task_id = task_id
        self.error = error



"""Melogen CLI tool."""

from __future__ import annotations

import json
import sys

import click

from melogenai.__version__ import __version__
from melogenai.client import Melogen
from melogenai.exceptions import MelogenError


def _get_client(api_key: str | None, base_url: str | None) -> Melogen:
    kwargs = {}
    if api_key:
        kwargs["api_key"] = api_key
    if base_url:
        kwargs["base_url"] = base_url
    return Melogen(**kwargs)


def _print_json(data: dict) -> None:
    click.echo(json.dumps(data, indent=2, ensure_ascii=False))


@click.group()
@click.version_option(__version__, prog_name="melogen")
@click.option("--api-key", envvar="MELOGEN_API_KEY", help="Melogen API key (or set MELOGEN_API_KEY)")
@click.option("--base-url", envvar="MELOGEN_BASE_URL", help="Override API base URL")
@click.pass_context
def cli(ctx: click.Context, api_key: str | None, base_url: str | None) -> None:
    """Melogen AI - AI-powered music tools."""
    ctx.ensure_object(dict)
    ctx.obj["api_key"] = api_key
    ctx.obj["base_url"] = base_url


@cli.command()
@click.argument("file_url")
@click.option("--format", "-f", "output_formats", multiple=True, help="Output format (midi, musicxml)")
@click.option("--no-midi", is_flag=True, help="Skip MIDI output")
@click.pass_context
def sheet2midi(ctx: click.Context, file_url: str, output_formats: tuple[str, ...], no_midi: bool) -> None:
    """Convert sheet music (PDF/PNG/JPG) to MIDI."""
    try:
        client = _get_client(ctx.obj["api_key"], ctx.obj["base_url"])
        click.echo(f"Submitting {file_url} for conversion...")

        formats = list(output_formats) if output_formats else None
        result = client.sheet2midi.convert(
            file_url, output_formats=formats, convert_midi=not no_midi
        )
        click.echo(f"Task {result.task_id} created. Waiting for completion...")

        final = client.sheet2midi.wait(result.task_id)
        click.echo(f"Status: {final.status}")
        if final.download_urls:
            click.echo("Download URLs:")
            for name, url in final.download_urls.items():
                click.echo(f"  {name}: {url}")
        else:
            _print_json(final.model_dump())
    except MelogenError as e:
        click.echo(f"Error: {e}", err=True)
        sys.exit(1)


@cli.command()
@click.argument("file_url")
@click.option("--stems", is_flag=True, help="Save separated instrument stems")
@click.option("--quantize", is_flag=True, help="Quantize MIDI output to nearest beat")
@click.pass_context
def music2midi(ctx: click.Context, file_url: str, stems: bool, quantize: bool) -> None:
    """Convert audio (MP3/WAV/FLAC/OGG/M4A) to MIDI."""
    try:
        client = _get_client(ctx.obj["api_key"], ctx.obj["base_url"])
        click.echo(f"Submitting {file_url} for conversion...")

        result = client.music2midi.convert(file_url, save_stems=stems, quantize=quantize)
        click.echo(f"Task {result.task_id} created. Waiting for completion...")

        final = client.music2midi.wait(result.task_id)
        click.echo(f"Status: {final.status}")
        if final.download_urls:
            click.echo("Download URLs:")
            for name, url in final.download_urls.items():
                click.echo(f"  {name}: {url}")
        else:
            _print_json(final.model_dump())
    except MelogenError as e:
        click.echo(f"Error: {e}", err=True)
        sys.exit(1)


@cli.command()
@click.option("--mxl", "mxl_file_url", help="URL of a MusicXML (.mxl) file")
@click.option("--image", "image_url", help="URL of a sheet music image")
@click.option("--pdf", "pdf_file_url", help="URL of a PDF score")
@click.option("--lang", default="en", help="Language for analysis (default: en)")
@click.option("--stream/--no-stream", default=True, help="Stream output (default: stream)")
@click.pass_context
def analysis(
    ctx: click.Context,
    mxl_file_url: str | None,
    image_url: str | None,
    pdf_file_url: str | None,
    lang: str,
    stream: bool,
) -> None:
    """Analyze music structure, tonality, harmony, and form."""
    if not any([mxl_file_url, image_url, pdf_file_url]):
        click.echo("Error: provide at least one of --mxl, --image, or --pdf", err=True)
        sys.exit(1)

    try:
        client = _get_client(ctx.obj["api_key"], ctx.obj["base_url"])
        kwargs = {"language": lang}
        if mxl_file_url:
            kwargs["mxl_file_url"] = mxl_file_url
        if image_url:
            kwargs["image_url"] = image_url
        if pdf_file_url:
            kwargs["pdf_file_url"] = pdf_file_url

        if stream:
            for chunk in client.analysis.analyze_stream(**kwargs):
                click.echo(chunk, nl=False)
            click.echo()
        else:
            result = client.analysis.analyze(**kwargs)
            click.echo(result.analysis)
    except MelogenError as e:
        click.echo(f"Error: {e}", err=True)
        sys.exit(1)


@cli.group()
def auth() -> None:
    """Authentication commands."""


@auth.command()
@click.pass_context
def verify(ctx: click.Context) -> None:
    """Verify your API key is valid."""
    try:
        client = _get_client(ctx.obj["api_key"], ctx.obj["base_url"])
        token_data = client._auth.get_token(client._config.api_key)
        click.echo(f"API key is valid. User ID: {token_data['user_id']}")
        if token_data.get("scopes"):
            click.echo(f"Scopes: {', '.join(token_data['scopes'])}")
    except MelogenError as e:
        click.echo(f"Error: {e}", err=True)
        sys.exit(1)
    except Exception as e:
        click.echo(f"Error: {e}", err=True)
        sys.exit(1)



"""Tests for analysis service client."""

from __future__ import annotations

import pytest
import respx

from melogenai.analysis.client import SUPPORTED_LANGUAGES
from melogenai.client import Melogen

from conftest import TEST_API_KEY, TEST_BASE_URL


class TestAnalysisSync:
    def test_analyze_with_image(self, mock_api: respx.MockRouter) -> None:
        sse_body = (
            "data: {\"content\": \"Key: C major. \"}\n\n"
            "data: {\"content\": \"Time: 4/4.\"}\n\n"
            "data: [DONE]\n\n"
        )
        mock_api.post("/api/analysis/analyze").respond(
            content=sse_body,
            content_type="text/event-stream",
        )

        client = Melogen(api_key=TEST_API_KEY, base_url=TEST_BASE_URL)
        result = client.analysis.analyze(image_url="https://example.com/score.png")

        assert result.status == "completed"
        assert "Key: C major" in result.analysis
        assert "Time: 4/4" in result.analysis

    def test_analyze_stream(self, mock_api: respx.MockRouter) -> None:
        sse_body = (
            "data: {\"content\": \"Part 1. \"}\n\n"
            "data: {\"content\": \"Part 2.\"}\n\n"
            "data: [DONE]\n\n"
        )
        mock_api.post("/api/analysis/analyze").respond(
            content=sse_body,
            content_type="text/event-stream",
        )

        client = Melogen(api_key=TEST_API_KEY, base_url=TEST_BASE_URL)
        chunks = list(client.analysis.analyze_stream(image_url="https://example.com/score.png"))

        assert len(chunks) == 2
        assert chunks[0] == "Part 1. "
        assert chunks[1] == "Part 2."

    def test_analyze_requires_input(self) -> None:
        client = Melogen(api_key=TEST_API_KEY, base_url=TEST_BASE_URL)
        with pytest.raises(ValueError, match="At least one of"):
            client.analysis.analyze()

    def test_analyze_rejects_bad_language(self) -> None:
        client = Melogen(api_key=TEST_API_KEY, base_url=TEST_BASE_URL)
        with pytest.raises(ValueError, match="Unsupported language"):
            client.analysis.analyze(image_url="https://example.com/score.png", language="xx")

    def test_supported_languages(self) -> None:
        assert "en" in SUPPORTED_LANGUAGES
        assert "zh" in SUPPORTED_LANGUAGES
        assert "ja" in SUPPORTED_LANGUAGES



"""Melogen AI Python SDK - AI-powered music tools."""

from melogenai.__version__ import __version__
from melogenai.client import AsyncMelogen, Melogen
from melogenai.exceptions import (
    AuthenticationError,
    InsufficientCreditsError,
    MelogenError,
    RateLimitError,
    TaskFailedError,
    TaskTimeoutError,
    ValidationError,
)

__all__ = [
    "__version__",
    "AsyncMelogen",
    "AuthenticationError",
    "InsufficientCreditsError",
    "Melogen",
    "MelogenError",
    "RateLimitError",
    "TaskFailedError",
    "TaskTimeoutError",
    "ValidationError",
]



"""API Key -> JWT token exchange with caching."""

from __future__ import annotations

import asyncio
import hashlib
import threading
import time
from typing import Any

import httpx

from melogenai.config import Config
from melogenai.exceptions import AuthenticationError

_MAX_CACHE_SIZE = 100
_TOKEN_CACHE_SECONDS = 3500  # ~58 min (JWT is 60 min)


class AuthManager:
    """Manages API Key -> JWT token exchange with thread-safe caching."""

    def __init__(self, config: Config):
        self._config = config
        self._cache: dict[str, dict[str, Any]] = {}
        self._sync_lock = threading.Lock()
        self._async_lock: asyncio.Lock | None = None

    @property
    def _alock(self) -> asyncio.Lock:
        if self._async_lock is None:
            self._async_lock = asyncio.Lock()
        return self._async_lock

    @staticmethod
    def _cache_key(api_key: str) -> str:
        return hashlib.sha256(api_key.encode()).hexdigest()

    def _evict_expired(self) -> None:
        now = time.time()
        expired = [k for k, v in self._cache.items() if v["expires_at"] <= now]
        for k in expired:
            del self._cache[k]

    def _get_cached(self, api_key: str) -> dict[str, Any] | None:
        key = self._cache_key(api_key)
        cached = self._cache.get(key)
        if cached and cached["expires_at"] > time.time():
            return cached
        return None

    def _store_token(self, api_key: str, data: dict[str, Any]) -> dict[str, Any]:
        key = self._cache_key(api_key)
        token_data = {
            "token": data["access_token"],
            "user_id": data["user_id"],
            "scopes": data.get("scopes"),
            "expires_at": time.time() + _TOKEN_CACHE_SECONDS,
        }
        self._evict_expired()
        if len(self._cache) >= _MAX_CACHE_SIZE:
            oldest_key = min(self._cache, key=lambda k: self._cache[k]["expires_at"])
            del self._cache[oldest_key]
        self._cache[key] = token_data
        return token_data

    def _parse_response(self, data: dict[str, Any]) -> dict[str, Any]:
        if not data.get("valid"):
            raise AuthenticationError("Invalid API key")
        return data

    def get_token(self, api_key: str) -> dict[str, Any]:
        """Exchange API key for a JWT token (sync)."""
        with self._sync_lock:
            cached = self._get_cached(api_key)
            if cached:
                return cached

            with httpx.Client(timeout=10) as client:
                resp = client.post(
                    f"{self._config.base_url}/api/api-keys/verify",
                    headers={"X-API-Key": api_key},
                )
                resp.raise_for_status()
                data = self._parse_response(resp.json())

            return self._store_token(api_key, data)

    async def async_get_token(self, api_key: str) -> dict[str, Any]:
        """Exchange API key for a JWT token (async)."""
        async with self._alock:
            cached = self._get_cached(api_key)
            if cached:
                return cached

            async with httpx.AsyncClient(timeout=10) as client:
                resp = await client.post(
                    f"{self._config.base_url}/api/api-keys/verify",
                    headers={"X-API-Key": api_key},
                )
                resp.raise_for_status()
                data = self._parse_response(resp.json())

            return self._store_token(api_key, data)

    def invalidate(self, api_key: str) -> None:
        """Remove cached token for an API key."""
        key = self._cache_key(api_key)
        self._cache.pop(key, None)
